// -*- C++ -*-

#pragma once

#include "cpu_functions.h"
#include "gettype.h"
#include "gpu_defs.h"
#include "lib/exception.h"
#include "lib/range.h"
#include "lib/stdinc.h"
#include "lib/vec.h"
#include "operators.h"

#define GOOPAX_CONSERVATIVE_ARITHMETIC_CONVERSIONS 0

/// \addtogroup types
/// gpu types, used in gpu kernels.
///

/// \addtogroup operators


namespace goopax {

template<typename T>
struct goopax_struct_type<T, typename std::conditional<true, void, typename T::goopax_struct_type>::type>
{
    using type = typename T::goopax_struct_type;
};


template<typename STR, typename X>
struct goopax_struct_changetype<STR, X, typename std::conditional<true, void, typename STR::template goopax_struct_changetype<X>>::type>
{
    using type = typename STR::template goopax_struct_changetype<X>;
};


namespace impl {


namespace interface {
#define MAKE_DELETE(T)                                                            \
    extern "C"                                                                    \
           GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_delete_##T(void* p) noexcept; \
    inline void goopax_delete(T* p)                                               \
    {                                                                             \
        goopax_4_delete_##T(p);                                                   \
    }

MAKE_DELETE(gpu_writer_base_impl)
MAKE_DELETE(kernel_impl) // FIXME: deprecated
MAKE_DELETE(loop_impl)
MAKE_DELETE(gpu_if_impl)
MAKE_DELETE(WELL512_impl)
MAKE_DELETE(vecWELL512lib_impl)
MAKE_DELETE(sha256_impl)
MAKE_DELETE(perfdata_math_entry_impl)
MAKE_DELETE(svm_buffer_impl)

extern "C"
       GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_delete_resource_base(void* p) noexcept;
inline void goopax_delete(resource_impl* p)
{
    goopax_4_delete_resource_base(p);
}


#undef MAKE_DELETE
#undef MAKE_INTRUSIVE

struct goopax_deleter
{
    template<typename T>
    void operator()(T* p)
    {
        goopax_delete(p);
    }
};

struct goopax_nodelete_deleter
{
    template<typename T>
    void operator()(T*)
    {
    }
};


extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_gpu_type_op1(unsigned int op, gpu_writer_base_impl* ret, const gpu_writer_base_impl* a);
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_gpu_type_op2(unsigned int op, gpu_writer_base_impl* ret, const gpu_writer_base_impl* a, const gpu_writer_base_impl* b);
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_gpu_type_op3(unsigned int op, gpu_writer_base_impl* ret, const gpu_writer_base_impl* a, const gpu_writer_base_impl* b, const gpu_writer_base_impl* c);
GOOPAX_DLL_EXPORT(4.0.0) gpu_writer_base_impl* goopax_4_gpu_type_new_impl() noexcept;
GOOPAX_DLL_EXPORT(4.0.0) gpu_writer_base_impl* goopax_4_gpu_type_uninit(unsigned int, memory::scope) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) gpu_writer_base_impl* goopax_4_gpu_type_assign(const void* b, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) gpu_writer_base_impl* goopax_4_gpu_type_convert(const gpu_writer_base_impl*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_loop_break(gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_loop_continue(gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_gpu_type_copyassign(gpu_writer_base_impl*, const gpu_writer_base_impl*) noexcept;
}
}


template<typename T, memory::scope SCOPE>
class local_scope_mem;
template<typename T, typename size_type, bool is_const>
class resource;
template<typename T, typename size_type>
class buffer;

template<typename T>
struct is_gpu_struct;

template<typename T, typename U>
struct is_gpu_struct_help : public false_type
{};

template<typename T>
struct is_gpu_struct_help<T, typename enable_if<(is_gpu_struct<typename goopax_struct_type<T>::type>::value || is_arithmetic<typename goopax_struct_type<T>::type>::value || is_pointer<typename goopax_struct_type<T>::type>::value || is_gpu_type<typename goopax_struct_type<T>::type>::value)>::type> : public integral_constant<bool, !is_arithmetic<T>::value && !is_pointer<T>::value>
{};


template<typename T>
struct is_gpu_struct : public is_gpu_struct_help<typename remove_cv<T>::type, void>
{};

template<typename T, memory::scope memscope>
struct is_gpu_struct<gpu_type<T, memscope>> : public false_type
{};


#if !GOOPAX_DISTRIB
void GOOPAX_DLL_EXPORT(4.0.0) goopax_4_debugcount_impl();
#endif

    inline void
    debugcount()
{
#if !GOOPAX_DISTRIB
    goopax_4_debugcount_impl();
#endif
}


template<bool is_unsigned, bool is_signed, unsigned int size>
struct type_enum_raw2;
template<>
struct type_enum_raw2<false, true, 8> : public integral_constant<unsigned int, 0>
{};
template<>
struct type_enum_raw2<false, true, 4> : public integral_constant<unsigned int, 1>
{};
template<>
struct type_enum_raw2<false, true, 2> : public integral_constant<unsigned int, 2>
{};
template<>
struct type_enum_raw2<false, true, 1> : public integral_constant<unsigned int, 3>
{};
template<>
struct type_enum_raw2<true, false, 8> : public integral_constant<unsigned int, 4>
{};
template<>
struct type_enum_raw2<true, false, 4> : public integral_constant<unsigned int, 5>
{};
template<>
struct type_enum_raw2<true, false, 2> : public integral_constant<unsigned int, 6>
{};
template<>
struct type_enum_raw2<true, false, 1> : public integral_constant<unsigned int, 7>
{};

template<typename T>
struct type_enum_raw : public type_enum_raw2<is_signed<T>::value, is_unsigned<T>::value, sizeof(T)>
{};

template<>
struct type_enum_raw<bool> : public integral_constant<unsigned int, 8>
{};
template<>
struct type_enum_raw<double> : public integral_constant<unsigned int, 9>
{};
template<>
struct type_enum_raw<float> : public integral_constant<unsigned int, 10>
{};
template<>
struct type_enum_raw<half> : public integral_constant<unsigned int, 11>
{};

template<typename T>
struct type_enum : public integral_constant<unsigned int, 2 * type_enum_raw<T>::value>
{};

template<typename T>
struct type_enum<checknan<T>> : public integral_constant<unsigned int, 2 * type_enum_raw<T>::value + 1>
{};

template<typename T>
struct type_enum<T*> : public integral_constant<unsigned int, 2 * (type_enum_raw<T>::value + 12)>
{};

template<typename T>
struct type_enum<checknan<T>*> : public integral_constant<unsigned int, 2 * (type_enum_raw<T>::value + 12) + 1>
{};


template<unsigned int N>
struct reverse_type_enum_raw;

template<>
struct reverse_type_enum_raw<0>
{
    using type = uint64_t;
};
template<>
struct reverse_type_enum_raw<1>
{
    using type = uint32_t;
};
template<>
struct reverse_type_enum_raw<2>
{
    using type = uint16_t;
};
template<>
struct reverse_type_enum_raw<3>
{
    using type = uint8_t;
};
template<>
struct reverse_type_enum_raw<4>
{
    using type = int64_t;
};
template<>
struct reverse_type_enum_raw<5>
{
    using type = int32_t;
};
template<>
struct reverse_type_enum_raw<6>
{
    using type = int16_t;
};
template<>
struct reverse_type_enum_raw<7>
{
    using type = int8_t;
};
template<>
struct reverse_type_enum_raw<8>
{
    using type = bool;
};
template<>
struct reverse_type_enum_raw<9>
{
    using type = double;
};
template<>
struct reverse_type_enum_raw<10>
{
    using type = float;
};
template<>
struct reverse_type_enum_raw<11>
{
    using type = half;
};

template<unsigned int N>
struct reverse_type_enum
{
    private:
    using type1 = typename reverse_type_enum_raw<(N / 2) % 12>::type;
    using type2 = typename conditional<N % 2 == 1, checknan<type1>, type1>::type;

    public:
    using type = typename conditional<((N / 2) >= 12), type2*, type2>::type;
};


using paramtype_t     = unsigned int;
using param_size_type = unsigned int;

/// \ingroup types
namespace typedefs {
using gpu_double    = gpu_type<double>;             ///< 64 bit floating point type
using gpu_float     = gpu_type<float>;              ///< 32 bit floatint point type
using gpu_half      = gpu_type<half>;               ///< 16 bit floatint point
using gpu_bool      = gpu_type<bool>;               ///< boolean
using gpu_char      = gpu_type<char>;               ///< character type
using gpu_schar     = gpu_type<signed char>;        ///< signed char
using gpu_uchar     = gpu_type<unsigned char>;      ///< unsigned char
using gpu_short     = gpu_type<short>;              ///< short integer, corresponding to CPU type 'short'
using gpu_int       = gpu_type<int>;                ///< integer
using gpu_long      = gpu_type<long>;               ///< long
using gpu_longlong  = gpu_type<long long>;          ///< long long
using gpu_ushort    = gpu_type<unsigned short>;     ///< unsigned short
using gpu_uint      = gpu_type<unsigned int>;       ///< unsigned int
using gpu_ulong     = gpu_type<unsigned long>;      ///< unsigned long
using gpu_ulonglong = gpu_type<unsigned long long>; ///< unsigned long long
using gpu_int64     = gpu_type<int64_t>;            ///< 64 bit integer
using gpu_int32     = gpu_type<int32_t>;            ///< 32 bit integer
using gpu_int16     = gpu_type<int16_t>;            ///< 16 bit integer
using gpu_int8      = gpu_type<int8_t>;             ///< 8 bit integer
using gpu_uint64    = gpu_type<uint64_t>;           ///< 64 bit unsigned int
using gpu_uint32    = gpu_type<uint32_t>;           ///< 32 bit unsigned int
using gpu_uint16    = gpu_type<uint16_t>;           ///< 16 bit unsigned int
using gpu_uint8     = gpu_type<uint8_t>;            ///< 8 bit unsigned int
using gpu_size_t    = gpu_type<size_t>;
using gpu_ssize_t   = gpu_type<std::intptr_t>;

static_assert(sizeof(size_t) == sizeof(void*), "");
}
using namespace typedefs;


template<typename T>
struct is_gpu_type : public false_type
{
};

template<typename T, memory::scope memscope>
struct is_gpu_type<gpu_type<T, memscope>> : public true_type
{};


// Sometimes getting alignment problems with std::unique_ptr. Using our own implementation here.
template<typename IMPL>
struct my_unique_ptr
{
    private:
    IMPL* impl;

    public:
    IMPL* get() const
    {
        return impl;
    }

    void reset(IMPL* ptr = nullptr)
    {
        *this = my_unique_ptr(ptr);
    }

    IMPL& operator*() const
    {
        return *get();
    }

    IMPL* operator->() const
    {
        return get();
    }

    my_unique_ptr(IMPL* ptr = nullptr)
        : impl(ptr)
    {}
    my_unique_ptr(const my_unique_ptr&) = delete;
    my_unique_ptr& operator=(const my_unique_ptr&) = delete;
    my_unique_ptr(my_unique_ptr&& b)
        : impl(b.impl)
    {
        b.impl = nullptr;
    }
    my_unique_ptr& operator=(my_unique_ptr&& b)
    {
        if (impl != nullptr)
        {
            goopax::impl::interface::goopax_deleter()(impl);
        }
        impl   = b.impl;
        b.impl = nullptr;
        return *this;
    }
    ~my_unique_ptr()
    {
        if (impl != nullptr)
        {
            goopax::impl::interface::goopax_deleter()(impl);
        }
    }
};


template<typename TYPE, typename IMPL>
struct wrapper
{
    my_unique_ptr<IMPL> impl;
};


/// \ingroup operators
/// conditional operator
template<typename T, memory::scope memscope>
inline gpu_type<T, memscope> cond(const gpu_bool& c, const gpu_type<T, memscope>& a, const gpu_type<T, memscope>& b);


struct INIT_IMPLNEW
{
};


template<typename A, typename B, bool fA, bool fB, bool lt, bool is_signed>
struct arith_result_type_help;

template<typename A, typename B, bool lt, bool is_signed>
struct arith_result_type_help<A, B, true, false, lt, is_signed>
{
    using type = A;
};
template<typename A, typename B, bool lt, bool is_signed>
struct arith_result_type_help<A, B, false, true, lt, is_signed>
{
    using type = B;
};

template<typename A, typename B>
struct arith_result_type_help<A, B, true, true, true, true>
{
    using type = B;
};
template<typename A, typename B>
struct arith_result_type_help<A, B, true, true, false, true>
{
    using type = A;
};

template<typename A, typename B>
struct arith_result_type_help<A, B, false, false, true, false>
{
    using type = B;
};
template<typename A, typename B>
struct arith_result_type_help<A, B, false, false, false, false>
{
    using type = A;
};

template<typename A, typename B, bool lt>
struct arith_result_type_help<A, B, false, false, lt, true>
{
    using type = typename make_signed<typename arith_result_type_help<A, B, false, false, lt, false>::type>::type;
};


template<typename A, typename B>
struct arith_result_type : public arith_result_type_help<A, B, is_floating_point<A>::value, is_floating_point<B>::value, (gettype<A>::size < gettype<B>::size), (is_signed<A>::value || is_signed<B>::value)>
{
    static_assert(!is_same<A, bool>::value, "");
    static_assert(!is_same<B, bool>::value, "");
};


/// \ingroup types
/// GPU data type, used in GPU kernels to store values in registers.
/// \tparam T corresponding CPU intrinsic type. Pointers are represented as gpu_type<T*> or gpu_type<const T*>, where T is a CPU intrinsic type or a goopax struct type.
template<typename T, memory::scope scope>
class gpu_type : public wrapper<gpu_type<T>, gpu_writer_base_impl>
{
    static_assert(is_integral<T>::value || is_floating_point<T>::value, "");
    static_assert(!std::is_const<T>::value, "Please use non-const, non-volatile types.");
    static_assert(!std::is_volatile<T>::value, "Please use non-const, non-volatile types.");
    static_assert(scope == memory::thread, "scope must be thread for non-pointers");

    static constexpr bool is_bool = is_same<T, bool>::value;
    static constexpr bool is_int  = is_integral<T>::value && !is_bool;


#define EIF(cond) template<typename EA = void, typename EB = typename enable_if<cond, EA>::type>

#define MAKE_OP0(PRE, RET, FUNC, ID)                                            \
    EIF(PRE)                                                                    \
    inline RET FUNC() const                                                     \
    {                                                                           \
        RET ret;                                                                \
        interface::goopax_4_gpu_type_op1(ID, ret.impl.get(), this->impl.get()); \
        return ret;                                                             \
    }
#define MAKE_OP1(PRE, RET, FUNC, ID, ARG1)                                                    \
    EIF(PRE)                                                                                  \
    inline RET FUNC(ARG1 b) const                                                             \
    {                                                                                         \
        RET ret;                                                                              \
        interface::goopax_4_gpu_type_op2(ID, ret.impl.get(), this->impl.get(), b.impl.get()); \
        return ret;                                                                           \
    }

    public:
    /// Compare two GPU values for equality
    MAKE_OP1(true, gpu_bool, operator==, 0, const gpu_type&)
    /// Compare two GPU values for inequality
    MAKE_OP1(true, gpu_bool, operator!=, 1, const gpu_type&)

    /// Compare two GPU values
    MAKE_OP1(!is_bool, gpu_bool, operator<, 2, const gpu_type&)
    /// Compare two GPU values
    MAKE_OP1(!is_bool, gpu_bool, operator>=, 3, const gpu_type&)
    /// boolean AND
    MAKE_OP1(is_bool, gpu_bool, operator&&, 4, const gpu_type&)
    /// boolean OR
    MAKE_OP1(is_bool, gpu_bool, operator||, 5, const gpu_type&)
    /// addition
    MAKE_OP1(!is_bool, gpu_type, operator+, 6, const gpu_type&)
    /// multiplication
    MAKE_OP1(!is_bool, gpu_type, operator*, 7, const gpu_type&)
    /// division
    MAKE_OP1(!is_bool, gpu_type, operator/, 8, const gpu_type&)
    /// bitwise AND
    MAKE_OP1(is_int, gpu_type, operator&, 9, const gpu_type&)
    /// bitwise OR
    MAKE_OP1(is_int, gpu_type, operator|, 10, const gpu_type&)
    /// bitwise xor
    MAKE_OP1(is_int, gpu_type, operator^, 11, const gpu_type&)
    /// left shift
    MAKE_OP1(!is_bool, gpu_type, operator<<, 12, const gpu_uint&)
    /// right shift
    MAKE_OP1(!is_bool, gpu_type, operator>>, 13, const gpu_uint&)
    /// modulo operator
    MAKE_OP1(is_int, gpu_type, operator%, 14, const gpu_type&)


#if GOOPAX_CONSERVATIVE_ARITHMETIC_CONVERSIONS
#define MAKESYM(OP)                                                                           \
    template<typename B, typename = typename enable_if<!is_same<T, B>::value>::type>          \
    gpu_type<decltype(*(T*)nullptr OP(*(B*)nullptr))> operator OP(const gpu_type<B>& b) const \
    {                                                                                         \
        static_assert(!is_same<T, B>::value, "");                                             \
        using WORKTYPE = gpu_type<decltype(*(T*)nullptr + (*(B*)nullptr))>;                   \
        return static_cast<WORKTYPE>(*this) OP(static_cast<WORKTYPE>(b));                     \
    }
#else
#define MAKESYM(OP)                                                                          \
    template<typename B, typename = typename enable_if<!is_same<T, B>::value>::type>         \
    gpu_type<typename arith_result_type<T, B>::type> operator OP(const gpu_type<B>& b) const \
    {                                                                                        \
        static_assert(!is_same<T, B>::value, "");                                            \
        using WORKTYPE = gpu_type<typename arith_result_type<T, B>::type>;                   \
        return static_cast<WORKTYPE>(*this) OP(static_cast<WORKTYPE>(b));                    \
    }
#endif

#define MAKESYMB(OP)                                                                 \
    template<typename B, typename = typename enable_if<!is_same<T, B>::value>::type> \
    gpu_bool operator OP(const gpu_type<B>& b) const                                 \
    {                                                                                \
        static_assert(!is_same<T, B>::value, "");                                    \
        using WORKTYPE = gpu_type<typename arith_result_type<T, B>::type>;           \
        return static_cast<WORKTYPE>(*this) OP(static_cast<WORKTYPE>(b));            \
    }

    MAKESYM(+)
    MAKESYM(-)
    MAKESYM(*)
    MAKESYM(/)
    MAKESYM(&)
    MAKESYM(|)
    MAKESYM(^)

    MAKESYMB(<)
    MAKESYMB(<=)
    MAKESYMB(>)
    MAKESYMB(>=)
    MAKESYMB(==)
    MAKESYMB(!=)

#undef MAKESYM
#undef MAKESYMB


    MAKE_OP0(is_int, gpu_type, operator~, 4)
    MAKE_OP0(is_bool, gpu_bool, operator!, 5)

    EIF(!is_bool)
    inline gpu_type operator-(const gpu_type& b) const
    {
        /// subtract
        return *this + (-b);
    }
    EIF(!is_bool)
    inline gpu_type operator-() const
    {
        /// unary -
        return *this * T(-1);
    }
    EIF(!is_bool)
    inline gpu_type operator+() const
    {
        /// unary +
        return *this;
    }
    EIF(!is_bool)
    inline gpu_bool operator>(const gpu_type& b) const
    {
        /// comparison
        return b < *this;
    }
    EIF(!is_bool)
    inline gpu_bool operator<=(const gpu_type& b) const
    {
        /// comparison
        return b >= *this;
    }


    EIF(is_int)
    inline gpu_type& operator++()
    {
        /// increment
        return *this += 1;
    }
    EIF(is_int)
    inline gpu_type& operator--()
    {
        /// decrement
        return *this -= 1;
    }
    EIF(is_int)
    inline gpu_type operator++(int)
    {
        /// increment
        gpu_type<T> ret = *this;
        ++*this;
        return ret;
    }
    EIF(is_int)
    inline gpu_type operator--(int)
    {
        /// decrement
        gpu_type<T> ret = *this;
        --*this;
        return ret;
    }
    friend inline gpu_type operator+(T a, const gpu_type& b) { return gpu_type(a) + b; }
    friend inline gpu_type operator-(T a, const gpu_type& b) { return gpu_type(a) - b; }
    friend inline gpu_type operator*(T a, const gpu_type& b) { return gpu_type(a) * b; }
    friend inline gpu_type operator/(T a, const gpu_type& b) { return gpu_type(a) / b; }
    friend inline gpu_bool operator<(T a, const gpu_type& b) { return gpu_type(a) < b; }
    friend inline gpu_bool operator<=(T a, const gpu_type& b) { return gpu_type(a) <= b; }
    friend inline gpu_bool operator>(T a, const gpu_type& b) { return gpu_type(a) > b; }
    friend inline gpu_bool operator>=(T a, const gpu_type& b) { return gpu_type(a) >= b; }
    friend inline gpu_bool operator==(T a, const gpu_type& b) { return gpu_type(a) == b; }
    friend inline gpu_bool operator==(const gpu_type& b, T a) { return gpu_type(a) == b; }
    friend inline gpu_bool operator!=(T a, const gpu_type& b) { return gpu_type(a) != b; }
    friend inline gpu_bool operator!=(const gpu_type& b, T a) { return gpu_type(a) != b; }
    friend inline gpu_type operator&&(T a, const gpu_type& b) { return gpu_type(a) && b; }
    friend inline gpu_type operator||(T a, const gpu_type& b) { return gpu_type(a) || b; }
    friend inline gpu_type operator&(T a, const gpu_type& b) { return gpu_type(a) & b; }
    friend inline gpu_type operator|(T a, const gpu_type& b) { return gpu_type(a) | b; }
    friend inline gpu_type operator^(T a, const gpu_type& b) { return gpu_type(a) ^ b; }
    friend inline gpu_type operator%(T a, const gpu_type& b) { return gpu_type(a) % b; }

    /// addition
    EIF(!is_bool)
    inline gpu_type& operator+=(const gpu_type& b) { return *this = *this + b; }
    /// subtraction
    EIF(!is_bool)
    inline gpu_type& operator-=(const gpu_type& b) { return *this = *this - b; }
    /// multiplication
    EIF(!is_bool)
    inline gpu_type& operator*=(const gpu_type& b) { return *this = *this * b; }
    /// division
    EIF(!is_bool)
    inline gpu_type& operator/=(const gpu_type& b) { return *this = *this / b; }
    /// bitwise AND
    EIF(is_int)
    inline gpu_type& operator&=(const gpu_type& b) { return *this = *this & b; }
    /// bitwise OR
    EIF(is_int)
    inline gpu_type& operator|=(const gpu_type& b) { return *this = *this | b; }
    /// bitwise XOR
    EIF(is_int)
    inline gpu_type& operator^=(const gpu_type& b) { return *this = *this ^ b; }
    /// left shift
    EIF(is_int)
    inline gpu_type& operator<<=(const gpu_uint& b) { return *this = *this << b; }
    /// right shift
    EIF(is_int)
    inline gpu_type& operator>>=(const gpu_uint& b) { return *this = *this >> b; }
    /// modulo
    EIF(is_int)
    inline gpu_type& operator%=(const gpu_type& b) { return *this = *this % b; }

#undef EIF
#undef MAKE_OP0
#undef MAKE_OP1

    void gpu_break() const
    {
        /// Break from loop. Only valid on loop variables. Calling gpu_break() on non-loop variables results in an error.
        check_errno(interface::goopax_4E_loop_break(this->impl.get()));
    }
    void gpu_continue() const
    {
        /// Continue loop. Only valid on loop variables. Calling gpu_continue() on non-loop variables results in an error.
        check_errno(interface::goopax_4E_loop_continue(this->impl.get()));
    }

    inline gpu_type()
        : gpu_type(INIT_IMPLNEW(), interface::goopax_4_gpu_type_uninit(type_enum<T>::value, scope))
    {
        /// default constructor for uninitialized variable.
    }

    template<typename U>
    explicit inline gpu_type(const struct wrapper<gpu_type<U>, gpu_writer_base_impl>& b)
        : gpu_type(INIT_IMPLNEW(), interface::goopax_4_gpu_type_convert(b.impl.get(), type_enum<T>::value))
    {
        /// explicit type conversion
    }

    inline gpu_type(T d)
        : gpu_type(INIT_IMPLNEW(), interface::goopax_4_gpu_type_assign(&d, type_enum<T>::value))
    {
        /// conversion from CPU type
    }

    template<typename U, typename = typename enable_if<is_half<T>::value && is_arithmetic<U>::value>::type>
    inline gpu_type(U d)
        : gpu_type((T)d)
    {
        /// conversion from CPU type
    }

    template<typename BT>
    inline gpu_type(const checknan<BT>& d)
        : gpu_type()
    {
        /// conversion from CPU debug type
        if (is_initialized(d))
            *this = unrange(d);
    }

    template<typename B, typename FOO = typename enable_if<(((is_int && is_integral<B>::value && !is_same<B, bool>::value) || (is_floating_point<T>::value && is_floating_point<B>::value)) && gettype<T>::size >= gettype<B>::size) || (is_floating_point<T>::value && is_integral<B>::value)>::type>

    inline gpu_type(const gpu_type<B>& b)
        : gpu_type(INIT_IMPLNEW(), interface::goopax_4_gpu_type_convert(b.impl.get(), type_enum<T>::value))
    {
        /// \brief Implicit type conversion
        /// The type conversion rules are stricter than the standard C/C++ type conversion rules to avoid unintended performance loss:
        /// Integer values are only implicitly converted to integers of at least the same size or to floating point numbers.
        /// floating point numbers are only implicitly converted to other floating point numbers of at least the same size.
        /// Other conversions need to be done explicitly, with static_cast, or C/C++ style cast.
    }

    inline gpu_type(const gpu_type& b)
        : gpu_type()
    {
        /// copy constructor
        *this = b;
    }
    inline gpu_type& operator=(const gpu_type& b)
    {
        /// assignment operator
        interface::goopax_4_gpu_type_copyassign(this->impl.get(), b.impl.get());
        return *this;
    }

    inline gpu_type(INIT_IMPLNEW, gpu_writer_base_impl* nt)
    {
        this->impl.reset(nt);
    }
};


template<typename RET, typename A>
RET make_op(unsigned int ID, const A& a)
{
    RET ret;
    interface::goopax_4_gpu_type_op1(ID, ret.impl.get(), a.impl.get());
    return ret;
}
template<typename RET, typename A, typename B>
RET make_op(unsigned int ID, const A& a, const B& b)
{
    RET ret;
    interface::goopax_4_gpu_type_op2(ID, ret.impl.get(), a.impl.get(), b.impl.get());
    return ret;
}
template<typename RET, typename A, typename B, typename C>
RET make_op(unsigned int ID, const A& a, const B& b, const C& c)
{
    RET ret;
    interface::goopax_4_gpu_type_op3(ID, ret.impl.get(), a.impl.get(), b.impl.get(), c.impl.get());
    return ret;
}

template<typename T, typename Enable = typename enable_if<is_floating_point<T>::value>::type>
inline gpu_bool isfinite(gpu_type<T> f)
{
    /// \ingroup Math
    /// returns true if f is finite
    return make_op<gpu_bool>(0, f);
}

template<typename T, typename Enable = typename enable_if<is_floating_point<T>::value>::type>
inline gpu_bool isinf(gpu_type<T> f)
{
    /// \ingroup Math
    /// returns true if f is infinite
    return make_op<gpu_bool>(1, f);
}

template<typename T, typename Enable = typename enable_if<is_floating_point<T>::value>::type>
inline gpu_bool isnan(gpu_type<T> f)
{
    /// \ingroup Math
    /// returns true if f is NaN
    return make_op<gpu_bool>(2, f);
}

template<typename T, typename Enable = typename enable_if<is_floating_point<T>::value>::type>
inline gpu_bool isnormal(gpu_type<T> f)
{
    /// \ingroup Math
    /// returns true if f is normal
    return make_op<gpu_bool>(3, f);
}


template<typename T, typename Enable>
gpu_type<T> min(gpu_type<T> a, gpu_type<T> b)
{
    /// \ingroup functions
    /// minimum of both numbers
    return make_op<gpu_type<T>>(15, a, b);
}

template<typename T, typename Enable>
gpu_type<T> max(gpu_type<T> a, gpu_type<T> b)
{
    /// \ingroup functions
    /// maximum of both numbers
    return make_op<gpu_type<T>>(16, a, b);
}

template<typename T>
gpu_type<T> min(initializer_list<gpu_type<T>> list)
{
    gpu_type<T> best = *list.begin();
    for (auto p = std::next(list.begin()); p != list.end(); ++p)
    {
        best = min(best, *p);
    }
    return best;
}

template<typename T>
gpu_type<T> max(initializer_list<gpu_type<T>> list)
{
    gpu_type<T> best = *list.begin();
    for (auto p = std::next(list.begin()); p != list.end(); ++p)
    {
        best = max(best, *p);
    }
    return best;
}


template<typename T>
gpu_type<T> clamp(gpu_type<T> a, typename enable_if<true, gpu_type<T>>::type b, typename enable_if<true, gpu_type<T>>::type c)
{
    /// \ingroup functions
    /// clamping a to range (b,c). Similar to max(min(a, c), b)
    return make_op<gpu_type<T>>(1, a, b, c);
}

template<typename T>
gpu_type<T> clamp(gpu_type<T> a, gpu_type<T> b, gpu_type<T> c)
{
    /// \ingroup functions
    /// clamping a to range (b,c). Similar to max(min(a, c), b)
    return make_op<gpu_type<T>>(1, a, b, c);
}


template<typename T>
gpu_type<T> min(gpu_type<T> a, typename unrangetype_core<T>::type b)
{
    /// \ingroup functions
    return min(a, gpu_type<T>(b));
}
template<typename T>
gpu_type<T> min(typename unrangetype_core<T>::type a, gpu_type<T> b)
{
    /// \ingroup functions
    return min(gpu_type<T>(a), b);
}

template<typename T>
gpu_type<T> max(gpu_type<T> a, typename unrangetype_core<T>::type b)
{
    /// \ingroup functions
    return max(a, gpu_type<T>(b));
}
template<typename T>
gpu_type<T> max(typename unrangetype_core<T>::type a, gpu_type<T> b)
{
    /// \ingroup functions
    return max(gpu_type<T>(a), b);
}


template<typename T, memory::scope memscope>
inline gpu_type<T, memscope> cond(const gpu_bool& c, const gpu_type<T, memscope>& a, const gpu_type<T, memscope>& b)
{
    /// \ingroup operators
    /// conditional function. Returns c ? a : b
    return make_op<gpu_type<T, memscope>>(0, c, a, b);
}
template<typename T>
gpu_type<T> cond(gpu_bool c, gpu_type<T> a, typename unrangetype_core<T>::type b)
{
    return cond(c, a, gpu_type<T>(b));
}
template<typename T>
gpu_type<T> cond(gpu_bool c, typename unrangetype_core<T>::type a, gpu_type<T> b)
{
    return cond(c, gpu_type<T>(a), b);
}
template<typename T, typename = typename enable_if<is_arithmetic<T>::value>::type>
typename gettype<T>::gpu cond(gpu_bool c, T a, T b)
{
    return cond(c, typename gettype<T>::gpu(a), typename gettype<T>::gpu(b));
}
template<typename T, typename = typename enable_if<is_arithmetic<T>::value>::type>
T cond(bool c, T a, T b)
{
    return c ? a : b;
}

template<typename T, typename = typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type>
gpu_type<typename make_unsigned<T>::type> abs(gpu_type<T> a)
{
    /// \ingroup functions
    /// \brief returns absolute value.
    /// if a is an integer, the return type is unsigned.
    return make_op<gpu_type<typename make_unsigned<T>::type>>(6, a);
}

template<typename T, typename = typename enable_if<is_floating_point<T>::value>::type>
gpu_type<T> abs(gpu_type<T> a)
{
    /// \ingroup functions
    /// returns absolute value
    return make_op<gpu_type<T>>(6, a);
}

template<typename T, typename = typename enable_if<is_floating_point<T>::value>::type>
gpu_type<T> floor(gpu_type<T> a)
{
    /// \ingroup Math
    /// round to -inf
    return make_op<gpu_type<T>>(7, a);
}

template<typename T, typename = typename enable_if<is_floating_point<T>::value>::type>
gpu_type<T> ceil(gpu_type<T> a)
{
    /// \ingroup Math
    /// round to +inf
    return make_op<gpu_type<T>>(8, a);
}

template<typename T, typename = typename enable_if<is_floating_point<T>::value>::type>
gpu_type<T> round(gpu_type<T> a)
{
    /// \ingroup Math
    /// round to nearest
    return make_op<gpu_type<T>>(9, a);
}


template<typename T, typename Enable = typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type>
gpu_uint countl_zero(gpu_type<T> a)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    return make_op<gpu_uint>(10, a);
}

template<typename T, typename Enable = typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type>
gpu_uint countl_one(gpu_type<T> a)
{
    return countl_zero(~a);
}


template<typename T, typename Enable = typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type>
[[deprecated("consider using countl_zero instead")]] gpu_uint clz(gpu_type<T> a)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    return countl_zero(a);
}


template<typename T, typename Enable = typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type>
gpu_uint popcount(gpu_type<T> a)
{
    /// \ingroup functions
    /// Counts the number of bits
    return make_op<gpu_uint>(11, a);
}

template<typename T, typename Enable = typename enable_if<is_integral<T>::value && !is_same<T, bool>::value>::type>
gpu_type<T> rotl(gpu_type<T> a, gpu_int b)
{
    /// \ingroup functions
    /// rotate left
    return make_op<gpu_type<T>>(17, a, b);
}

template<typename T, typename Enable = typename enable_if<is_integral<T>::value && !is_same<T, bool>::value>::type>
gpu_type<T> rotr(gpu_type<T> a, gpu_int b)
{
    /// \ingroup functions
    /// rotate right
    return rotl(a, -b);
}


#define MAKE_ARITH(OP)                                                                                                            \
    template<typename A, typename B, typename = typename enable_if<(is_floating_point<A>::value && is_integral<B>::value)>::type> \
    inline typename gettype<A>::gpu operator OP(const A& a, const gpu_type<B>& b)                                                 \
    {                                                                                                                             \
        return typename gettype<A>::gpu(a) OP typename gettype<A>::gpu(b);                                                        \
    }                                                                                                                             \
    template<typename A, typename B, typename = typename enable_if<(is_integral<A>::value && is_floating_point<B>::value)>::type> \
    inline typename gettype<B>::gpu operator OP(const gpu_type<A>& a, const B& b)                                                 \
    {                                                                                                                             \
        return typename gettype<B>::gpu(a) OP typename gettype<B>::gpu(b);                                                        \
    }

#define MAKE_FUNC(FUNC)                                                                                                           \
    template<typename A, typename B, typename = typename enable_if<(is_floating_point<A>::value && is_integral<B>::value)>::type> \
    inline typename gettype<A>::gpu FUNC(const A& a, const gpu_type<B>& b)                                                        \
    {                                                                                                                             \
        return FUNC(typename gettype<A>::gpu(a), b);                                                                              \
    }                                                                                                                             \
    template<typename A, typename B, typename = typename enable_if<(is_integral<A>::value && is_floating_point<B>::value)>::type> \
    inline typename gettype<B>::gpu FUNC(const gpu_type<A>& a, const B& b)                                                        \
    {                                                                                                                             \
        return FUNC(a, typename gettype<B>::gpu(b));                                                                              \
    }

#define MAKE_FUNC3(FUNC, ARG1)                                                                                                    \
    template<typename A, typename B, typename = typename enable_if<(is_floating_point<A>::value && is_integral<B>::value)>::type> \
    inline typename gettype<A>::gpu FUNC(ARG1 c, const A& a, const gpu_type<B>& b)                                                \
    {                                                                                                                             \
        return FUNC(c, typename gettype<A>::gpu(a), b);                                                                           \
    }                                                                                                                             \
    template<typename A, typename B, typename = typename enable_if<(is_integral<A>::value && is_floating_point<B>::value)>::type> \
    inline typename gettype<B>::gpu FUNC(ARG1 c, const gpu_type<A>& a, const B& b)                                                \
    {                                                                                                                             \
        return FUNC(c, a, typename gettype<B>::gpu(b));                                                                           \
    }

MAKE_ARITH(+)
MAKE_ARITH(-)
MAKE_ARITH(*)
MAKE_ARITH(/)
MAKE_FUNC(max)
MAKE_FUNC(min)
MAKE_FUNC3(cond, gpu_bool)

#undef MAKE_ARITH
#undef MAKE_FUNC
#undef MAKE_FUNC3

template<typename A, typename B, typename ENABLE = typename enable_if<is_integral<A>::value>::type>
gpu_type<typename unrangetype_core<A>::type> operator<<(A a, gpu_type<B> b)
{
    return gpu_type<typename unrangetype_core<A>::type>(a) << b;
}
template<typename A, typename B, typename ENABLE = typename enable_if<is_integral<A>::value>::type>
gpu_type<typename unrangetype_core<A>::type> operator>>(A a, gpu_type<B> b)
{
    return gpu_type<typename unrangetype_core<A>::type>(a) >> b;
}


namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_thread_local_id(gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_thread_group_id(gpu_writer_base_impl*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_thread(unsigned int*, int) noexcept;
}
}


template<typename T, typename>
struct get_X_type
{
    using type = typename get_X_type<typename goopax_struct_type<typename remove_cv<T>::type>::type>::type;
};

template<typename T>
struct get_X_type<T, typename enable_if<(is_arithmetic<T>::value || is_pointer<T>::value) && !is_checknan<T>::value>::type>
{
    using type = xtype_int;
};
template<>
struct get_X_type<xtype_int>
{
    using type = xtype_int;
};
template<typename T>
struct get_X_type<checknan<T>>
{
    using type = checknan<xtype_int>;
};
template<typename T, memory::scope memscope>
struct get_X_type<gpu_type<T, memscope>>
{
    using type = xtype_gpu;
};
template<>
struct get_X_type<void>
{
    using type = void;
};
template<typename T>
struct get_X_type<gpu_struct::struct_item_map<T>>
{
    using type = gpu_struct::struct_item_map<typename get_X_type<T>::type>;
};
template<typename T, size_t N>
struct get_X_type<gpu_struct::struct_mod_test<T, N>>
{
    using type = gpu_struct::struct_mod_test<typename get_X_type<T>::type, N>;
};
template<typename T, unsigned int POW>
struct get_X_type<gpu_struct::struct_type_probe<T, POW>>
{
    using type = gpu_struct::struct_type_probe<typename get_X_type<T>::type, POW>;
};
template<typename T, typename TO>
struct get_X_type<gpu_struct::struct_change_type<T, TO>>
{
    using type = gpu_struct::struct_change_type<typename get_X_type<T>::type, TO>;
};
}
using impl::cond;


using impl::gettype;

using impl::is_gpu_type;

using namespace impl::typedefs;

using impl::clamp;
using impl::clz;
using impl::cond;
using impl::countl_one;
using impl::countl_zero;
using impl::popcount;
using impl::rotl;
using impl::rotr;

using impl::abs;
using impl::ceil;
using impl::floor;
using impl::isfinite;
using impl::isinf;
using impl::isnan;
using impl::isnormal;
using impl::max;
using impl::min;
using impl::round;

}


namespace std {
/// numeric_limits overload for GPU types
template<typename T>
struct numeric_limits<goopax::gpu_type<T>> : std::numeric_limits<T>
{};

template<typename T>
struct make_signed<goopax::gpu_type<T>>
{
    using type = goopax::gpu_type<typename make_signed<T>::type>;
};

template<typename T>
struct make_unsigned<goopax::gpu_type<T>>
{
    using type = goopax::gpu_type<typename make_unsigned<T>::type>;
};


using goopax::abs;
using goopax::ceil;
using goopax::floor;
using goopax::isfinite;
using goopax::isinf;
using goopax::isnan;
using goopax::isnormal;
using goopax::max;
using goopax::min;
using goopax::round;
#if __cplusplus >= 201700
using goopax::clamp;
#endif
}
