// -*- C++ -*-


#pragma once

#include "../proto.h"
#include "./defs.h"
#include "stdinc.h"
#include "vec.h"
#include <atomic>
#if __cplusplus >= 202000
#if __has_include(<bit>)
#include <bit>
#endif
#endif
#include <cstdint>
#include <sstream>

#include "operators.h"
#include "stdinc.h"


namespace goopax {
namespace impl {
template<typename T>
struct is_gpu_struct;
}
}


namespace goopax {
namespace impl {
template<typename STR>
struct output_goopax_struct_helper;
template<typename T>
output_goopax_struct_helper<T> output_goopax_struct(const T& t);
}
}


namespace goopax {
template<typename T, typename ENABLE = void>
struct goopax_struct_type;
template<class STR, class X, class ENABLE = void>
struct goopax_struct_changetype;

}


namespace goopax {

namespace impl {
enum init_state
{
    init_yes        = 1,
    init_no         = 2,
    init_destructed = 3
};


namespace interface {
#if __cplusplus >= 201700
using demangle_deletefunc_t = void(const char*) noexcept;
#else
using demangle_deletefunc_t = void(const char*);
#endif
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) const char* goopax_4_type_demangle_gnuc(const char*, demangle_deletefunc_t**) noexcept;
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_range_throw_failure(init_state init, const char* tiname) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) const char* goopax_4_ex_what() noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_io_warn(const char*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) void goopax_4_check_construct(const void*) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) void goopax_4_check_destruct(const void*) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) void goopax_4_check_read(const void*) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) init_state goopax_4_check_read_pair(const void*) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) void goopax_4_check_write(const void*, init_state) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) init_state goopax_4_check_atomic_read(const void*) noexcept;
GOOPAX_DLL_EXPORT(4.2.0) void goopax_4_check_atomic_write(const void*, init_state) noexcept;
}
}


inline string pretty_typename(const char* tiname)
{
    interface::demangle_deletefunc_t* deleter;
    const char*                       tmp = interface::goopax_4_type_demangle_gnuc(tiname, &deleter);
    string                            ret;
    if (tmp == nullptr)
    {
        ret = "[unknown]";
    }
    else
    {
        ret = tmp;
        deleter(tmp);
    }
    return ret;
}

inline string pretty_typename(const type_info& ti)
{
    return pretty_typename(ti.name());
}


template<typename T>
struct is_checknan : public false_type
{};

template<typename T>
struct is_checknan<checknan<T>> : public true_type
{};


template<typename T>
struct unrangetype_core
{
    using type = T;
};
template<typename T>
struct unrangetype_core<checknan<T>>
{
    typedef T type;
};


template<typename T>
struct unrangetype;

template<typename T, typename = void>
struct unrangetype_help;

template<typename T>
struct unrangetype_help<T, typename enable_if<is_arithmetic<T>::value || is_pointer<T>::value>::type>
{
    using type = T;
};


template<typename T>
struct unrangetype : public unrangetype_help<T>
{
};

template<typename T>
struct unrangetype<checknan<T>>
{
    typedef T type;
};
template<typename T>
struct unrangetype<goopax::impl::gpu_type<T>>
{
    using type = goopax::impl::gpu_type<T>;
};

template<typename T, size_t N>
struct unrangetype<std::array<T, N>>
{
    using type = std::array<typename unrangetype<T>::type, N>;
};

template<typename A, typename B>
struct unrangetype<pair<A, B>>
{
    using type = pair<typename unrangetype<A>::type, typename unrangetype<B>::type>;
};


template<typename T, size_t N>
inline std::array<typename unrangetype<T>::type, N> unrange(const std::array<T, N>& a);
template<typename T>
inline
    typename conditional<is_base_of<typename unrangetype<T>::type, T>::value, const typename unrangetype<T>::type&, typename unrangetype<T>::type>::type
    unrange(const T& a);


template<typename T, size_t N>
inline std::array<typename unrangetype<T>::type, N> unrange(const std::array<T, N>& a)
{
    std::array<typename unrangetype<T>::type, N> ret;
    for (size_t k = 0; k < N; ++k)
        ret[k] = unrange(a[k]);
    return ret;
}

template<typename T>
inline
    typename conditional<is_base_of<typename unrangetype<T>::type, T>::value, const typename unrangetype<T>::type&, typename unrangetype<T>::type>::type
    unrange(const T& a)
{
    return a;
}

#if GOOPAX_DEBUG_FOR_CL
template<typename T>
struct cldebug_unrangetype : public unrangetype<T>
{};
#else
template<typename T>
struct cldebug_unrangetype
{
    typedef T type;
};
#endif


template<typename T, typename = void>
struct is_single_number : false_type
{};

template<typename T>
struct is_single_number<T,
                        typename enable_if<is_fundamental<typename unrangetype_core<T>::type>::value || is_same<typename unrangetype<T>::type, half>::value>::type> : true_type
{};

template<typename T>
struct is_single_number<goopax::impl::gpu_type<T>> : true_type
{};


namespace interface {
template<typename = void>
struct callmap
{
    static callmap data;
    void (*barrier)(int, unsigned int, unsigned int, void*) noexcept = nullptr;

    void (*check_construct)(const void*) noexcept                = nullptr;
    void (*check_destruct)(const void*) noexcept                 = nullptr;
    void (*check_read)(const void*) noexcept                     = nullptr;
    init_state (*check_read_pair)(const void*) noexcept          = nullptr;
    void (*check_write)(const void*, init_state) noexcept        = nullptr;
    init_state (*check_atomic_read)(const void*) noexcept        = nullptr;
    void (*check_atomic_write)(const void*, init_state) noexcept = nullptr;

    void (*ta_inc_fbt)() noexcept                                                                                                                = nullptr;
    void (*threadcall)(void (*)(unsigned int, unsigned int, void* const* const, void*), unsigned int, unsigned int, void* const* const) noexcept = nullptr;
    void (*dumpstream_print)(const checknan<unsigned int>*, const unsigned int) noexcept                                                         = nullptr;
    void (*io_warn)(const char*, unsigned int) noexcept                                                                                          = nullptr;
};
template<typename T>
callmap<T> callmap<T>::data;
}

struct warning
{
    unique_ptr<stringstream> ss;
    unsigned int             MODE = 0;

    warning()
        : ss(new stringstream())
    {}
    warning& operator=(warning&&) = delete;
    inline ~warning()
    {
        if (ss.get() != nullptr)
        {
            const string sss = ss->str();
#ifdef GOOPAX_CPU_CALL
            interface::callmap<>::data.io_warn(sss.c_str(), MODE);
#else
            interface::goopax_4_io_warn(sss.c_str(), MODE);
#endif
        }
    }

    template<class T>
    friend warning&& operator<<(warning&& s, const T& t)
    {
        using std::operator<<;
        static_cast<ostream&>(*s.ss) << t;
        return move(s);
    }
    friend warning&& operator<<(warning&& s, ostream& (*func)(ostream&))
    {
        using std::operator<<;
        static_cast<ostream&>(*s.ss) << func;
        return move(s);
    }
};

struct exwhat : public warning
{
    exwhat()
    {
        this->MODE = 1;
    }
};


struct checknan_base
{
    //    volatile init_state init;

    inline checknan_base()
    {
#ifdef GOOPAX_CPU_CALL
        interface::callmap<>::data.check_construct(this);
#else
        interface::goopax_4_check_construct(this);
#endif
    }

    inline ~checknan_base()
    {
#ifdef GOOPAX_CPU_CALL
        interface::callmap<>::data.check_destruct(this);
#else
        interface::goopax_4_check_destruct(this);
#endif
    }
    checknan_base(const checknan_base&) = default;
    checknan_base& operator=(const checknan_base&) = default;
};


template<typename T>
class checknan : private checknan_base
{
    static_assert(!is_checknan<T>::value, "");

    template<typename U>
    friend class checknan;

#ifdef GOOPAX_CPU_CALL
    public:
#endif

    T data;

    std::pair<T, init_state> read_pair() const
    {
        static_assert(sizeof(checknan) == sizeof(T), "");

#ifdef GOOPAX_CPU_CALL
        init_state init = interface::callmap<>::data.check_read_pair(this);
#else
        init_state init = interface::goopax_4_check_read_pair(this);
#endif
        return { data, init };
    }

    T read() const
    {
        static_assert(sizeof(checknan) == sizeof(T), "");

#ifdef GOOPAX_CPU_CALL
        interface::callmap<>::data.check_read(this);
#else
        interface::goopax_4_check_read(this);
#endif
        return data;
    }

    void write_pair(std::pair<T, init_state> value)
    {
#ifdef GOOPAX_CPU_CALL
        interface::callmap<>::data.check_write(this, value.second);
#else
        interface::goopax_4_check_write(this, value.second);
#endif
        data = value.first;
    }

    void write(T value)
    {
        write_pair({ value, init_yes });
    }

    checknan<T> atomic_read()
    {
#ifdef GOOPAX_CPU_CALL
        init_state init = interface::callmap<>::data.check_atomic_read(this);
#else
        init_state init = interface::goopax_4_check_atomic_read(this);
#endif

        checknan<T> ret;
        ret.write_pair({ data, init });
        return ret;
    }

    void atomic_write(checknan<T> a)
    {
        auto ap = a.read_pair();
#ifdef GOOPAX_CPU_CALL
        interface::callmap<>::data.check_atomic_write(this, ap.second);
#else
        interface::goopax_4_check_atomic_write(this, ap.second);
#endif
        data = ap.first;
    }


    public:
    inline bool is_uninitialized() const
    {
        return (read_pair().second == init_no);
    }
    inline bool is_initialized() const
    {
        return (read_pair().second == init_yes);
    }

    inline T& unrange()
    {
        read();
        return data;
    }

    inline operator const T&() const
    {
        read();
        return data;
    }

    inline checknan& operator++()
    {
        *this = *this + 1;
        return *this;
    }

    inline checknan& operator--()
    {
        *this = *this - 1;
        return *this;
    }

    inline checknan operator++(int)
    {
        const checknan oldval = *this;
        ++*this;
        return oldval;
    }

    inline checknan operator--(int)
    {
        const checknan oldval = *this;
        --*this;
        return oldval;
    }


    template<typename DUMMY = int>
    inline checknan& operator|=(const checknan& b)
    {
        *this = *this | goopax::impl::unrange(b);
        return *this;
    }

    template<typename DUMMY = int>
    inline checknan& operator&=(const checknan& b)
    {
        *this = *this & goopax::impl::unrange(b);
        return *this;
    }

    template<typename DUMMY = int>
    inline checknan& operator^=(const checknan& b)
    {
        *this = *this ^ goopax::impl::unrange(b);
        return *this;
    }

    template<typename DUMMY = int>
    inline checknan& operator>>=(const unsigned int b)
    {
        *this = *this >> b;
        return *this;
    }

    template<typename DUMMY = int>
    inline checknan& operator<<=(const unsigned int b)
    {
        *this = *this << b;
        return *this;
    }


    template<typename U>
    inline checknan& operator=(const checknan<U>& in)
    {
        write_pair(in.read_pair());
        return *this;
    }
    inline checknan& operator=(const checknan& in)
    {
        write_pair(in.read_pair());
        return *this;
    }


    template<typename U>
    inline checknan(const checknan<U>& in)
    {
        *this = in;
    }

    inline checknan(const checknan& in)
    {
        *this = in;
    }

    inline checknan(T in)
    {
        write(in);
    }

    template<typename U = void, typename = typename std::enable_if<std::is_same<T, half_float::half>::value, U>::type>
    inline checknan(half_float::detail::expr a)
        : checknan((half_float::half)a)
    {}

    inline checknan()
    {}

    template<typename U>
    checknan& operator=(const U& in)
    {
        write(in);
        return *this;
    }


    friend std::ostream& operator<<(std::ostream& s, const checknan& n)
    {
        auto p = n.read_pair();
        if (p.second == init_yes)
            s << p.first;
        else if (p.second == init_no)
            s << "UNINITIALIZED";
        else if (p.second == init_destructed)
            s << "DESTRUCTED";
        else
            s << "UNCONSTRUCTED";
        return s;
    }
    /*
      friend ostream& operator << (stringstream& s, const checknan& n)
      {
	return static_cast<ostream&>(s) << n;
      }
      */

    friend std::istream& operator>>(std::istream& s, checknan& n)
    {
        T ret;
        s >> ret;
        if (s)
            n = ret;
        else
            n = checknan();
        return s;
    }


    template<typename B>
    checknan& operator+=(const B& b)
    {
        return *this = this->read() + b;
    }

    template<typename B>
    checknan& operator-=(const B& b)
    {
        return *this = this->read() - b;
    }

    template<typename B>
    checknan& operator*=(const B& b)
    {
        return *this = this->read() * b;
    }

    template<typename B>
    checknan& operator/=(const B& b)
    {
        return *this = this->read() / b;
    }
    template<typename B, typename = typename std::enable_if<std::is_integral<B>::value>::type>
    checknan& operator%=(const B& b)
    {
        return *this = this->read() % b;
    }
};


template<typename T>
inline T& unrange(checknan<T>& a)
{
    return a.unrange();
}

template<typename A, typename B>
struct unrange_opcheck : public integral_constant<bool, (!is_same<A, typename unrangetype<A>::type>::value || !is_same<B, typename unrangetype<B>::type>::value) && !is_base_of<istream, A>::value && !is_base_of<ostream, A>::value>
{};

#define RANGE_DEF_OP(OP)                                                                                \
    template<typename A, typename B, typename = typename enable_if<unrange_opcheck<A, B>::value>::type> \
    auto operator OP(const A& a, const B& b)                                                            \
        ->typename std::remove_reference<decltype(unrange(a) OP unrange(b))>::type                      \
    {                                                                                                   \
        return unrange(a) OP unrange(b);                                                                \
    }

#define RANGE_DEF_FUNC(FUNC)                                                                                 \
    template<typename A, typename B, typename = typename std::enable_if<unrange_opcheck<A, B>::value>::type> \
    auto FUNC(const A& a, const B& b)                                                                        \
        ->typename std::remove_reference<decltype(FUNC(unrange(a), unrange(b)))>::type                       \
    {                                                                                                        \
        return FUNC(unrange(a), unrange(b));                                                                 \
    }

RANGE_DEF_FUNC(max)
RANGE_DEF_FUNC(min)

#undef RANGE_DEF_OP
#undef RANGE_DEF_FUNC

template<typename T>
inline decltype(~T()) operator~(const checknan<T>& a)
{
    return ~unrange(a);
}

template<typename T, typename SIGNED>
inline T abshelp(T a, true_type, SIGNED)
{
    return std::abs(a);
}
template<typename T>
inline T abshelp(T a, false_type, true_type)
{
    return std::abs(a);
}
template<typename T>
inline T abshelp(T a, false_type, false_type)
{
    return a;
}


template<typename T>
inline T abs(checknan<T> a)
{
    return abshelp(unrange(a), std::is_floating_point<T>(), std::is_signed<T>());
}

#define MAKEFUNC(FUNC)                                                                               \
    template<typename T>                                                                             \
    inline T FUNC(checknan<T> a)                                                                     \
    {                                                                                                \
        return std::FUNC((typename conditional<is_same<T, half>::value, float, T>::type)unrange(a)); \
    }

MAKEFUNC(floor)
MAKEFUNC(ceil)
MAKEFUNC(round)
MAKEFUNC(isfinite)
MAKEFUNC(isnormal)
MAKEFUNC(isnan)
MAKEFUNC(isinf)

GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)


#undef MAKEFUNC


namespace checkrange_types {
namespace debug {
template<typename T>
using checknantype = checknan<typename unrangetype<typename remove_const<typename remove_reference<T>::type>::type>::type>;
template<typename T>
struct checknantypestr
{
    typedef checknantype<T> type;
};

using Tlong_double  = checknantype<long double>;
using Tdouble       = checknantype<double>;
using Tfloat        = checknantype<float>;
using Thalf         = checknantype<half>;
using Tbool         = checknantype<bool>;
using Tchar         = checknantype<char>;
using Tschar        = checknantype<signed char>;
using Tuchar        = checknantype<unsigned char>;
using Tshort        = checknantype<short>;
using Tint          = checknantype<int>;
using Tlong         = checknantype<long>;
using Tlonglong     = checknantype<long long>;
using Tushort       = checknantype<unsigned short>;
using Tuint         = checknantype<unsigned int>;
using Tulong        = checknantype<unsigned long>;
using Tulonglong    = checknantype<unsigned long long>;
using Tint8_t       = checknantype<int8_t>;
using Tint16_t      = checknantype<int16_t>;
using Tint32_t      = checknantype<int32_t>;
using Tint64_t      = checknantype<int64_t>;
using Tuint8_t      = checknantype<uint8_t>;
using Tuint16_t     = checknantype<uint16_t>;
using Tuint32_t     = checknantype<uint32_t>;
using Tuint64_t     = checknantype<uint64_t>;
using Tsize_t       = checknantype<size_t>;
using Tssize_t      = checknantype<typename make_signed<size_t>::type>;
using Tunsigned_int = checknantype<unsigned int>;
}
namespace release {
template<typename T>
using checknantype = typename unrangetype<typename remove_const<typename remove_reference<T>::type>::type>::type;
template<typename T>
struct checknantypestr
{
    typedef checknantype<T> type;
};

using Tlong_double  = long double;
using Tdouble       = double;
using Tfloat        = float;
using Thalf         = half;
using Tbool         = bool;
using Tchar         = char;
using Tschar        = signed char;
using Tuchar        = unsigned char;
using Tshort        = short;
using Tint          = int;
using Tlong         = long;
using Tlonglong     = long long;
using Tushort       = unsigned short;
using Tuint         = unsigned int;
using Tulong        = unsigned long;
using Tulonglong    = unsigned long long;
using Tint8_t       = int8_t;
using Tint16_t      = int16_t;
using Tint32_t      = int32_t;
using Tint64_t      = int64_t;
using Tuint8_t      = uint8_t;
using Tuint16_t     = uint16_t;
using Tuint32_t     = uint32_t;
using Tuint64_t     = uint64_t;
using Tsize_t       = size_t;
using Tssize_t      = typename make_signed<size_t>::type;
using Tunsigned_int = unsigned int;
}
}


template<typename T>
inline bool is_uninitialized(const T&)
{
    return false;
}
template<typename T>
inline bool is_uninitialized(const checknan<T>& a)
{
    return a.is_uninitialized();
}
template<typename T>
inline bool is_initialized(const T&)
{
    return true;
}
template<typename T>
inline bool is_initialized(const checknan<T>& a)
{
    return a.is_initialized();
}

template<typename T>
struct is_half : is_same<typename unrangetype_core<T>::type, half>
{};
}

using impl::checknan;

using impl::is_checknan;
using impl::is_initialized;
using impl::is_uninitialized;
using impl::unrange;
using impl::unrangetype;
using impl::unrangetype_core;

using impl::half;
using impl::is_half;
using impl::output_goopax_struct;
}


namespace std {
using goopax::impl::abs;
using goopax::impl::ceil;
using goopax::impl::floor;
using goopax::impl::isfinite;
using goopax::impl::isinf;
using goopax::impl::isnan;
using goopax::impl::isnormal;
using goopax::impl::max;
using goopax::impl::min;
using goopax::impl::round;


template<typename T>
struct numeric_limits<goopax::checknan<T>> : public numeric_limits<T>
{
};


template<typename T>
struct make_signed<goopax::checknan<T>>
{
    using type = goopax::checknan<typename make_signed<T>::type>;
};

template<typename T>
struct make_unsigned<goopax::checknan<T>>
{
    using type = goopax::checknan<typename make_unsigned<T>::type>;
};
}

namespace goopax {
namespace release {
namespace types {
using namespace goopax::impl::checkrange_types::release;
}
}
namespace debug {
namespace types {
using namespace goopax::impl::checkrange_types::debug;
}
}
}
