// -*- C++ -*-

#pragma once

#include "lib/defs.h"
#include <cstddef>
#include <type_traits>


namespace goopax {
struct memory
{
    enum scope
    {
        thread,
        threadgroup,
        device,
        none
    };
};


namespace impl {
namespace machinery {
template<typename BUF>
struct pas_buf;
}

enum errno_type
{
    SUCCESS = 0,
    ERR_general,
    ERR_integrity,
    ERR_syntax,
    ERR_memory,
    ERR_assertion,
    ERR_backend,
    ERR_user
};


template<typename T, memory::scope scope = (std::is_pointer<T>::value ? memory::device : memory::thread)>
class gpu_type;
template<typename T, typename SIZE_TYPE>
class buffer;
template<typename T, typename SIZE_TYPE, bool is_const>
class resource;


template<typename T>
struct image_type_info;
template<unsigned int DIM, typename T, bool normalized = !image_type_info<T>::allow_unnormalized>
class image_buffer;
template<unsigned int DIM, typename T, bool normalized = !image_type_info<T>::allow_unnormalized>
class image_array_buffer;
template<unsigned int DIM, typename T, bool normalized = !image_type_info<T>::allow_unnormalized>
class image_resource;
template<unsigned int DIM, typename T, bool normalized = !image_type_info<T>::allow_unnormalized>
class image_array_resource;

template<typename T>
struct gather_result;


struct gpu_writer_base_impl;

namespace gpu_struct {
struct mapframe_base;
}

template<typename T>
struct is_gpu_type;
template<typename T, typename = void>
struct get_X_type;

template<typename T>
struct is_gather_type;


struct xtype_int
{
};
using xtype_gpu  = gpu_type<unsigned int>;
using xtype_void = xtype_int;
using std::size_t;

struct kernel_impl;
struct texbuf;
struct loop_impl;
struct resource_impl;
struct gpu_if_impl;
struct WELL512_impl;
struct vecWELL512lib_impl;
struct sha256_impl;
struct perfdata_math_entry_impl;
class svm_buffer_impl;
template<typename T, typename EIF = void>
struct gettype;
template<typename T>
class goopax_future;

template<typename G, typename T>
typename gettype<T>::gpu work_group_reduce(T value, unsigned int width);

namespace gpu_struct {
struct mapframe_base;
template<class STR>
struct mapframe;
template<typename T>
struct struct_item_map;
template<typename T, size_t N>
struct struct_mod_test;
template<typename T, unsigned int POW>
struct struct_type_probe;
template<typename T, typename TO>
struct struct_change_type;

}


inline void check_errno(errno_type err);
namespace gpu_thread {
inline unsigned int local_size();
}


}
}


#define GOOPAX_INSTANTIATE_MATH_ARG(FUNC) \
    FUNC(sin)                             \
    FUNC(sinpi)                           \
    FUNC(cos)                             \
    FUNC(cospi)                           \
    FUNC(tan)                             \
    FUNC(tanpi)                           \
    FUNC(exp)                             \
    FUNC(exp2)                            \
    FUNC(log)                             \
    FUNC(log2)                            \
    FUNC(asin)                            \
    FUNC(asinpi)                          \
    FUNC(acos)                            \
    FUNC(acospi)                          \
    FUNC(atan)                            \
    FUNC(atanpi)                          \
    FUNC(sinh)                            \
    FUNC(cosh)                            \
    FUNC(tanh)                            \
    FUNC(asinh)                           \
    FUNC(acosh)                           \
    FUNC(atanh)                           \
    FUNC(erf)                             \
    FUNC(erfc)                            \
    FUNC(tgamma)


#define GOOPAX_INSTANTIATE_MATH_WRAP(FUNC) \
    FUNC(sqrt)                             \
    FUNC(cbrt)


#define GOOPAX_INSTANTIATE_MATH(FUNC) \
    GOOPAX_INSTANTIATE_MATH_ARG(FUNC)
//  GOOPAX_INSTANTIATE_MATH_ARG_INT(FUNC)


namespace half_float {
class half;
}

namespace goopax {

namespace impl {
template<typename T>
class checknan;
#define MAKEFUNC(FUNC)   \
    template<typename T> \
    inline T FUNC(goopax::impl::checknan<T> a);

GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)
#undef MAKEFUNC

template<typename T>
inline decltype(~T()) operator~(const checknan<T>& a);

}
}

namespace goopax {
namespace impl {

#define MAKEFUNC(FUNC)                                                                                \
    inline goopax::impl::gpu_type<half_float::half> FUNC(goopax::impl::gpu_type<half_float::half> a); \
    inline goopax::impl::gpu_type<float>            FUNC(goopax::impl::gpu_type<float> a);            \
    inline goopax::impl::gpu_type<double>           FUNC(goopax::impl::gpu_type<double> a);


GOOPAX_INSTANTIATE_MATH(MAKEFUNC)
GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)
#undef MAKEFUNC
}
}

namespace std {
#define MAKEFUNC(FUNC) using goopax::impl::FUNC;

GOOPAX_INSTANTIATE_MATH(MAKEFUNC)
GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)

#undef MAKEFUNC
}


namespace goopax {
using impl::gpu_type;
}
